@@ -54,14 +54,6 @@ using namespace KTextEditor;
 namespace Python
 {
 
-ContextBuilder::ContextBuilder()
-    : m_mapAst(true)
-    , m_prebuilding(false)
-    , m_indentInformationCache(0)
-{
-
-}
-
 ReferencedTopDUContext ContextBuilder::build(const IndexedString& url, Ast* node, ReferencedTopDUContext updateContext)
 {
     if (!updateContext) {
@@ -81,7 +73,6 @@ ReferencedTopDUContext ContextBuilder::build(const IndexedString& url, Ast* node
     } else {
         qDebug() << " ====> DUCHAIN ====>     building duchain for" << url.str();
     }
-    m_isScheduledForReparsing = false;
     return ContextBuilderBase::build(url, node, updateContext);
 }
 
@@ -95,6 +86,16 @@ IndexedString ContextBuilder::currentlyParsedDocument() const
     return m_currentlyParsedDocument;
 }
 
+void ContextBuilder::setCurrentlyParsedDocument(const IndexedString &document)
+{
+    m_currentlyParsedDocument = document;
+}
+
+void ContextBuilder::setFutureModificationRevision(const ModificationRevision &rev)
+{
+    m_futureModificationRevision = rev;
+}
+
 RangeInRevision ContextBuilder::rangeForNode(Ast* node, bool moveRight)
 {
     return RangeInRevision(node->startLine, node->startCol, node->endLine, node->endCol + (int) moveRight);
@@ -123,6 +124,16 @@ DUContext* ContextBuilder::newContext(const RangeInRevision& range)
     return new PythonNormalDUContext(range, currentContext());
 }
 
+void ContextBuilder::addUnresolvedImport(const IndexedString &module)
+{
+    m_unresolvedImports.append(module);
+}
+
+QList<IndexedString> ContextBuilder::unresolvedImports() const
+{
+    return m_unresolvedImports;
+}
+
 void ContextBuilder::setEditor(PythonEditorIntegrator* editor)
 {
     m_editor = editor;
@@ -291,12 +302,7 @@ void ContextBuilder::visitCode(CodeAst* node) {
     Q_ASSERT(currentlyParsedDocument().toUrl().isValid());
     if ( currentlyParsedDocument() != doc ) {
         // Search for the python built-in functions file, and dump its contents into the current file.
-        TopDUContext* internal = 0;
-        {
-            DUChainReadLocker lock(DUChain::lock());
-            internal = DUChain::self()->chainForDocument(doc); // TODO add startup-check and error message, this must exist
-        }
-        
+        auto internal = Helper::getDocumentationFileContext();
         if ( ! internal ) {
             // If the built-in functions file is not yet parsed, schedule it with a high priority.
             m_unresolvedImports.append(doc);
@@ -308,9 +314,8 @@ void ContextBuilder::visitCode(CodeAst* node) {
             // KDevelop::ICore::self()->languageController()->backgroundParser()->parseDocuments();
         }
         else {
-            DUChainWriteLocker wlock(DUChain::lock());
+            DUChainWriteLocker wlock;
             currentContext()->addImportedParentContext(internal);
-            m_builtinFunctionsContext = TopDUContextPointer(internal);
         }
     }
     AstDefaultVisitor::visitCode(node);
@@ -357,31 +362,33 @@ QPair<KUrl, QStringList> ContextBuilder::findModulePath(const QString& name, con
                 leftNameComponents.removeFirst();
             }
             QString testFilename = tmp.path(KUrl::AddTrailingSlash) + component;
-            KUrl sourceUrl = testFilename + ".py";
+            tmp.cd(component);
+            QFileInfo sourcedir(testFilename);
+
             // we can only parse those, so we don't care about anything else for now.
             // Any C modules (.so, .dll) will be ignored, and highlighted as "not found". TODO fix this
-            QFile sourcefile(testFilename + ".py");
-            QFileInfo sourcedir(testFilename);
-            tmp.cd(component);
-            if ( ! sourcedir.exists() || ! sourcedir.isDir() || leftNameComponents.isEmpty() ) {
-                // If the search cannot continue further down into a hierarchy of directories,
-                // the file matching the next name component will be returned,
-                // toegether with a list of names which must be resolved inside that file.
-                if ( sourcefile.exists() ) {
-                    sourceUrl.cleanPath();
-                    return QPair<KUrl, QStringList>(sourceUrl, leftNameComponents);
+            static QStringList valid_extensions{".py", ".pyx"};
+            for(auto extension: valid_extensions) {
+                QFile sourcefile(testFilename + extension);
+                if ( ! sourcedir.exists() || ! sourcedir.isDir() || leftNameComponents.isEmpty() ) {
+                    // If the search cannot continue further down into a hierarchy of directories,
+                    // the file matching the next name component will be returned,
+                    // toegether with a list of names which must be resolved inside that file.
+                    if ( sourcefile.exists() ) {
+                        KUrl sourceUrl = testFilename + extension;
+                        sourceUrl.cleanPath();
+                        return qMakePair(sourceUrl, leftNameComponents);
+                    }
+                    else if ( sourcedir.exists() && sourcedir.isDir() ) {
+                        KUrl path(testFilename + "/__init__.py");
+                        path.cleanPath();
+                        return qMakePair(path, leftNameComponents);
+                    }
                 }
-                else if ( sourcedir.exists() && sourcedir.isDir() ) {
-                    KUrl path(testFilename + "/__init__.py");
-                    path.cleanPath();
-                    return QPair<KUrl, QStringList>(path, leftNameComponents);
-                }
-                kDebug() << "RESULT:" << "No module path found.";
-                break;
             }
         }
     }
-    return QPair<KUrl, QStringList>(KUrl(), QStringList());
+    return {};
 }
 
 RangeInRevision ContextBuilder::rangeForArgumentsContext(FunctionDefinitionAst* node)
